/*
 * =====================================================================================
 *
 *       Filename:  button.cpp
 *
 *    Description:  Button class
 *
 *        Version:  1.0
 *        Created:  15.05.2016 14:35:48
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Simon Ferber (https://www.ist-schlau.de), simon@ist-schlau.de
 *   Organization:  Kiwi Homebrew Computer
 *
 * =====================================================================================
 */

#include "button.h"
#include "screen.h"

/* Public parts */

button::button (uint16_t x, uint16_t y, uint16_t w, uint16_t h, std::string btn_caption)
{
    positionX = x;
    positionY = y;
    width = w;
    height = h;
    caption = btn_caption;
    widgetType = type::Button;
    pressed = false;
    callback = nullptr;
    render();
}

button::~button()
{
}

void button::render (bool pressed)
{
    uint16_t color1, color2;

    if (pressed)
    {
        color1 = 0x3333;
        color2 = 0x6666;
    }
    else
    {
        color1 = 0x6666;
        color2 = 0x3333;
    }

    // Draw button background
    VDP_BOX btn_background, caption_box;
    uint16_t btn_bgcolor=0x5555;
    btn_background.left = positionX;
    btn_background.top = positionY;
    btn_background.width = width;
    btn_background.height = height;
    vdp_critical |= 0x01;
    v9990_SetCmdWriteMask (0xffff);
    VDPWriteReg (VDP_LOP,VDP_LOP_WCSC);
    v9990_DrawFilledBox (&btn_background, btn_bgcolor);
    wait_vdp();
    // Draw button border
    v9990_DrawLine (positionX, positionY, width, 0, color1); // top border
    v9990_DrawLine (positionX, positionY, 0, height, color1); // left border
    v9990_DrawLine (positionX, positionY + height, width, 0, color2); // bottom border
    v9990_DrawLine (positionX + width, positionY, 0, height, color2); // right border
    // Draw button caption (centered)
    // character size 8x8
    uint16_t capX = positionX + ((width - (caption.size() << 3)) >> 1);   // horizontal center
    uint16_t capY = positionY + ((height - 8) >> 1) + 1;   // vertical center
    caption_box.left = capX;
    caption_box.top = capY;
    caption_box.width = 8;
    caption_box.height = 8;
    VDPWriteReg (VDP_FC,0x0000); // Foreground color
    WRITE_VDP (VDP_REGDAT,0x0000); // Foreground color
    VDPWriteReg (VDP_BC,btn_bgcolor); // Background color
    WRITE_VDP (VDP_REGDAT,btn_bgcolor); //Background color
    v9990_SetCmdWriteMask (0xffff);
    wait_vdp();

    for (std::string::iterator it=caption.begin(); it!=caption.end(); ++it)
    {
        v9990_CopyVRamCharToXY (&caption_box, VRAM_FONT + ((*it-32)<<3)); // -32 because font position != ASCII code
        caption_box.left += 8;
    }
    vdp_critical &= ~0x01;
}

void button::doAction (widget::action wAction)
{
    if (wAction == action::Selected)
    {
        render (true);
    }
    else
    {
        if (wAction == action::None)
        {
            if (lastAction == action::Selected)
            {
                render (false);
            }
        }
    }

    lastAction = wAction;
}

widget::action button::getlastAction()
{
    return lastAction;
}

