/*
 * =====================================================================================
 *
 *       Filename:  hamterm.c
 *
 *    Description:  Hamterm main()
 *
 *        Version:  1.0
 *        Created:  12.05.2016 21:29:03
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Simon Ferber (https://www.ist-schlau.de), simon@ist-schlau.de
 *   Organization:  Kiwi Homebrew Computer
 *
 * =====================================================================================
 */

#include "screen.h"
#include "input.h"
#include "interrupt.h"
#include "button.h"
#include "label.h"
#include "textbox.h"
#include "textinput.h"
#include "hammacro.h"
#include "splash.h"

#include <string>


void FindReplace (std::string& line, std::string oldString, std::string newString)
{
    const size_t oldSize = oldString.length();

    /* do nothing if line is shorter than the string to find */
    if (oldSize > line.length())
    {
        return;
    }

    const size_t newSize = newString.length();

    for (size_t pos = 0; ; pos += newSize)
    {
        /* Locate the substring to replace */
        pos = line.find (oldString, pos);

        if (pos == std::string::npos)
        {
            return;
        }

        if (oldSize == newSize)
        {
            /* if they're same size, use std::string::replace */
            line.replace (pos, oldSize, newString);
        }
        else
        {
            /* if not same size, replace by erasing and inserting */
            line.erase (pos, oldSize);
            line.insert (pos, newString);
        }
    }
}

void sendmacro(uint8_t macronum, macro *m, textbox *dest)
{
    /*
    if(macronum==1) {
        dest->cputc(0x1b);
        dest->cputc('L');
        return;
    }
    */
    std::string macrostr;
    macrostr = *m->getMacro(macronum);
    FindReplace(macrostr, (std::string) "<CALL>", ((textinput *)(dest->getTopScreen()->getWidget(1)))->getValue());
    FindReplace(macrostr, (std::string) "<NAME>", ((textinput *)(dest->getTopScreen()->getWidget(3)))->getValue());
    for (std::string::iterator it=macrostr.begin(); it!=macrostr.end(); ++it)
    {
    	dest->cputc(*it);
    	if(*it == '\n')
    		dest->cputc('\r');
    }
    dest->cputc('\r');
    dest->cputc('\n');
}

void callback1 (widget *tb_input)
{
    sendmacro(0, (*tb_input).getTopScreen()->macros, dynamic_cast <textbox*>(tb_input->getTopScreen()->getWidget(13))); // getWidget(widget number)
}

void callback2 (widget *tb_input)
{
    sendmacro(1, tb_input->getTopScreen()->macros, dynamic_cast <textbox*>(tb_input->getTopScreen()->getWidget(13))); // getWidget(widget number)
}

void callback3 (widget *tb_input)
{
    sendmacro(2, tb_input->getTopScreen()->macros, dynamic_cast <textbox*>(tb_input->getTopScreen()->getWidget(13))); // getWidget(widget number)
}

void callback4 (widget *tb_input)
{
    sendmacro(3, tb_input->getTopScreen()->macros, dynamic_cast <textbox*>(tb_input->getTopScreen()->getWidget(13))); // getWidget(widget number)
}

void callback5 (widget *tb_input)
{
    sendmacro(4, tb_input->getTopScreen()->macros, dynamic_cast <textbox*>(tb_input->getTopScreen()->getWidget(13))); // getWidget(widget number)
}

void callback6 (widget *tb_input)
{
    sendmacro(5, tb_input->getTopScreen()->macros, dynamic_cast <textbox*>(tb_input->getTopScreen()->getWidget(13))); // getWidget(widget number)
}

void callback7 (widget *tb_input)
{
    sendmacro(6, tb_input->getTopScreen()->macros, dynamic_cast <textbox*>(tb_input->getTopScreen()->getWidget(13))); // getWidget(widget number)
}

int main()
{
    /* New screen object:
     * set resolution and color table
     * upload font to vram */
    screen myScreen;

    showSplash(43, 40); /* (positionX, positionY) */

    macro fmacro;
    myScreen.macros = &fmacro;

    myScreen.clearScreen();

    /* Textbox output */
    textbox tb_output (1, 15, 509, 100);

    /* Label/Input call */
    label lbl_call (2, 121, (std::string) "Call");
    textinput tinpt_call (38, 118, 76, 11);
    //textbox tinpt_call (38, 118, 82, 11);

    /* Label/Input QTH */
    label lbl_QTH (124, 121, (std::string) "QTH");
    textinput tinpt_QTH (152, 118, 124, 11);
    //textbox tinpt_QTH (152, 118, 124, 12);

    /* Label/Input OP */
    label lbl_OP (282, 121, (std::string) "OP");
    textinput tinpt_OP (302, 118, 124, 11);
    //textbox tinpt_OP (302, 118, 124, 12);

    /* Label/Input Freq */
    label lbl_FREQ (432, 121, (std::string) "Freq");
    textinput tinpt_FREQ (468, 118, 40, 11);
    //textbox tinpt_FREQ (468, 118, 40, 12);

    /* Buttons - parameters: posX, posY, width, height, caption */
    button btn_CQ (2, 133, 60, 12, (std::string) "CQ");
    button btn_Answer (66, 133, 60, 12, (std::string) "Answer");
    button btn_QSO (130, 133, 60, 12, (std::string) "QSO");
    button btn_KN (194, 133, 60, 12, (std::string) "KN");
    button btn_SK (258, 133, 60, 12, (std::string) "SK");
    button btn_Me (322, 133, 60, 12, (std::string) "Me/QTH");
    button btn_Brag (386, 133, 60, 12, (std::string) "Brag");
    button btn_Clear (450, 133, 60, 12, (std::string) "Clear");

    /* Textbox input */
    textbox tb_input (1, 148, 509, 62);

    /* Add all widgets to screen object */
    myScreen.addWidget (&tb_output); /* Widget number 0 */
    myScreen.addWidget (&tinpt_call); /* Widget number 1 */
    myScreen.addWidget (&tinpt_QTH); /* Widget number 2 */
    myScreen.addWidget (&tinpt_OP); /* Widget number 3 */
    myScreen.addWidget (&tinpt_FREQ); /* Widget number 4 */
    myScreen.addWidget (&btn_CQ); /* Widget number 5 */
    myScreen.addWidget (&btn_Answer); /* Widget number 6 */
    myScreen.addWidget (&btn_QSO); /* Widget number 7 */
    myScreen.addWidget (&btn_KN); /* Widget number 8 */
    myScreen.addWidget (&btn_SK); /* Widget number 9 */
    myScreen.addWidget (&btn_Me); /* Widget number 10 */
    myScreen.addWidget (&btn_Brag); /* Widget number 11 */
    myScreen.addWidget (&btn_Clear); /* Widget number 12 */
    myScreen.addWidget (&tb_input); /* Widget number 13 */

    /* Setup callbacks for button clicks */
    btn_CQ.setCallback (&callback1);
    btn_Answer.setCallback (&callback2);
    btn_QSO.setCallback (&callback3);
    btn_KN.setCallback (&callback4);
    btn_SK.setCallback (&callback5);
    btn_Me.setCallback (&callback6);
    btn_Brag.setCallback (&callback7);
    btn_Clear.setCallback (&callback2);

    /* mouse/keyboard/rs232 */
    input myInput;
    myInput.setRS232();

    /* Initialize interrupts */
    interrupt myIRR (&myScreen);
    myIRR.setISR (interrupt::ISRduart, 6); /* keyboard, mouse, rsr232 in/output */
    myIRR.setISR (interrupt::ISRtimer, 5); /* timer interrupt */
    myIRR.initTimer();
    myIRR.setLevel (4);
    WRITEREG (DUART_IMR,0x22); /*  interrupt on RxRDYA & RxRDYB */

    /*
    std::string caption = "Hallo, das ist ein Test!\n\r\33v";

    for (std::string::iterator it=caption.begin(); it!=caption.end(); ++it)
    {
        tb_input.cputc (*it);
    }

    std::string caption2 = "Und noch ein anderer Test!\n\r";

    for (std::string::iterator it=caption2.begin(); it!=caption2.end(); ++it)
    {
        tb_output.cputc (*it);
        tb_input.cputc (*it);
    }
    */

    /* main loop */
    while (1)
    {
        myInput.dispatch(); // better call from some timer ISR?
        myInput.MouseAction (&myScreen);

        if (myInput.rs232getbuffcnt())
        {
            unsigned char c;
            c = myInput.rs232getchar();
            tb_output.cputc (c);
        }

        if (myInput.kbgetbuffcnt())
        {
            uint32_t c;
            uint8_t tab;
            tab = 0;
            c = myInput.kbgetchar();

            /* Check for scancodes F1-F10 (see any Atari ST documentation) */
            if ( (0xff & (c >> 16) > 58) && (0xff & (c >> 16) < 70) )
            {
                sendmacro((c>>16)-59, &fmacro, &tb_input);
            }

            if ( (0xff & c) == 0x09) /* TAB key */
            {
                tab = 1;
                for (uint8_t i=myScreen.getActiveWidget()->getWidgetNumber()+1; i < myScreen.getMyWidgets().size(); i++)
                {
                    if( (myScreen.getWidget(i)->getType() == widget::type::Textinput) || (myScreen.getWidget(i)->getType() == widget::type::Textbox) )
                    {
                        myScreen.getWidget(i)->doAction(widget::action::Selected);
                        tab++;
                        break;
                    }
                }
                if ( tab == 1)
                {
                    for (uint8_t i=0; i < myScreen.getActiveWidget()->getWidgetNumber()-1; i++)
                    {
                        if( (myScreen.getWidget(i)->getType() == widget::type::Textinput) || (myScreen.getWidget(i)->getType() == widget::type::Textbox) )
                        {
                            myScreen.getWidget(i)->doAction(widget::action::Selected);
                            break;
                        }
                    }
                }
            }

            /* if there is an active widget (textbox/input) print the character */
            if (myScreen.getActiveWidget() != nullptr && !tab)
            {
                /* Translate arrow key scancodes to VT52 ESC codes */
                if(0xff & (c>>16) == 72) /* Arrow up */
                {
                        ((textbox *)myScreen.getActiveWidget())->cputc (0x1b); /* ESC */
                        ((textbox *)myScreen.getActiveWidget())->cputc ('A'); /* UP */
                }
                if(0xff & (c>>16) == 80) /* Arrow down */
                {
                        ((textbox *)myScreen.getActiveWidget())->cputc (0x1b); /* ESC */
                        ((textbox *)myScreen.getActiveWidget())->cputc ('B'); /* DOWN */
                }
                if(0xff & (c>>16) == 75) /* Arrow left */
                {
                        ((textbox *)myScreen.getActiveWidget())->cputc (0x1b); /* ESC */
                        ((textbox *)myScreen.getActiveWidget())->cputc ('D'); /* LEFT */
                }
                if(0xff & (c>>16) == 77) /* Arrow right */
                {
                        ((textbox *)myScreen.getActiveWidget())->cputc (0x1b); /* ESC */
                        ((textbox *)myScreen.getActiveWidget())->cputc ('C'); /* RIGHT */
                }

                if(myScreen.getActiveWidget()->getType() == widget::type::Textinput)
                {
                    ((textinput *)myScreen.getActiveWidget())->cputc (c);
                } else
                {
                    ((textbox *)myScreen.getActiveWidget())->cputc (c);

                    if ((uint8_t) c == '\r')
                    {
                        ((textbox *)myScreen.getActiveWidget())->cputc ('\n');
                    }
                }
            }
        }
    }

    return 0;
}
