/*
 * =====================================================================================
 *
 *       Filename:  input.cpp
 *
 *    Description:  Input class (keyboard / mouse)
 *
 *        Version:  1.0
 *        Created:  16.05.2016 09:12:08
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Simon Ferber (https://www.ist-schlau.de), simon@ist-schlau.de
 *   Organization:  Kiwi Homebrew Computer
 *
 * =====================================================================================
 */

#include "input.h"
#include "screen.h"
#include "mpointer.h"
#include "keyb_de.h"

#define MAGIC_MOUSE 0x81

//uint8_t input::shifty;
//uint8_t input::kb_ticks;
//uint16_t input::kb_initial;
//uint16_t input::kb_repeat;
//int input::kb_dead;
//int input::kb_altnum;
//long input::kb_last_key;
//struct keytbl {
//    /* 128-sized array giving char codes for each scan code */
//    const uint8_t *norm;
//    const uint8_t *shft;
//    const uint8_t *caps;
//    /* couples of (scan code, char code), ended by byte zero */
//    const uint8_t *altnorm;
//    const uint8_t *altshft;
//    const uint8_t *altcaps;
//    /* table of at most eight dead key translation tables */
//    const uint8_t * const *dead;
//};

#define set_sr(a)                         \
__extension__                             \
({uint16_t _r, _a = (a);                     \
  __asm__ volatile                        \
  ("move.w %%sr,%0\n\t"                     \
   "move.w %1,%%sr"                         \
  : "=&d"(_r)       /*  outputs */        \
  : "nd"(_a)        /*  inputs  */        \
  : "cc", "memory"   /*  clobbered */      \
  );                                      \
  _r;                                     \
})

/* rs232 buffer */
unsigned char input::rs232buffer[RS232_BUFF_SIZE];
volatile unsigned char *input::rs232inpt, *input::rs232outpt; // Ringbuffer rs232
volatile uint16_t input::rs232buffcnt;

/* Mouse/Keyboard buffer */
unsigned char input::mskbbuffer[BUFF_SIZE];
volatile unsigned char *input::mskbinpt, *input::mskboutpt; // Ringbuffer mouse/keyboard
volatile unsigned char input::mskbbuffcnt;

volatile uint8_t duart_critical;

input::input()
{
    loadMousepointer();
    rs232inpt = rs232buffer;
    rs232outpt = rs232buffer;
    rs232buffcnt = 0;
    mskbinpt = mskbbuffer;
    mskboutpt = mskbbuffer;
    kbinpt = kbbuffer;
    kboutpt = kbbuffer;
    msinpt = msbuffer;
    msoutpt = msbuffer;
//struct input::keytbl input::current_keytbl;
    current_keytbl.norm  = keytbl_de_norm; 
    current_keytbl.shft = keytbl_de_shft;
    current_keytbl.caps  = keytbl_de_caps;
    current_keytbl.altnorm = keytbl_de_altnorm;
    current_keytbl.altshft = keytbl_de_altshft;
    current_keytbl.altcaps = keytbl_de_altcaps;
    //current_keytbl.dead  = 
    duart_critical = 0;
}

void input::put_rs232buff (unsigned char c)
{
    if (rs232buffcnt<RS232_BUFF_SIZE) // If buffer not full
    {
        *rs232inpt = c; // Put character into buffer
        rs232inpt++; // Increment pointer
        rs232buffcnt++;

        if (rs232inpt >= (rs232buffer + RS232_BUFF_SIZE)) // Pointer wrapping
        {
            rs232inpt = rs232buffer;
        }
    }
}

void input::put_mskbbuff (unsigned char c)
{
    if (mskbbuffcnt<BUFF_SIZE) // If buffer not full
    {
        *mskbinpt = c; // Put character into buffer
        mskbinpt++; // Increment pointer
        mskbbuffcnt++;

        if (mskbinpt >= (mskbbuffer + BUFF_SIZE)) // Pointer wrapping
        {
            mskbinpt = mskbbuffer;
        }
    }
}

//void input::put_kbbuff (unsigned char c)
void input::put_kbbuff (uint32_t c)
{
    if (kbbuffcnt<BUFF_SIZE) // If buffer not full
    {
        *kbinpt = c; // Put character into buffer
        kbinpt++; // Increment pointer
        kbbuffcnt++;

        if (kbinpt >= kbbuffer + BUFF_SIZE) // Pointer wrapping
        {
            kbinpt = kbbuffer;
        }
    }
}

void input::put_msbuff (unsigned char c)
{
    if (msbuffcnt<BUFF_SIZE) // If buffer not full
    {
        *msinpt = c; // Put character into buffer
        msinpt++; // Increment pointer
        msbuffcnt++;

        if (msinpt >= msbuffer + BUFF_SIZE) // Pointer wrapping
        {
            msinpt = msbuffer;
        }
    }
}

unsigned char input::rs232getchar()
{
    /* When calling, make sure buffer is not empty!
     * check rs232buffcnt > 0 */
    unsigned char byte=0;
    static uint16_t old_sr;
    //asm volatile ("move.w   #0x2700, %sr"); // all interrupts off
    old_sr = set_sr(0x2700);
    byte = *rs232outpt; // Get byte
    rs232buffcnt--; // Decrement buffer count
    rs232outpt++; // Increment pointer
    set_sr(old_sr);
    //asm volatile ("move.w   #0x2500, %sr"); // all interrupts off

    if (rs232outpt >= rs232buffer + RS232_BUFF_SIZE) // Pointer wrapping
    {
        rs232outpt = rs232buffer;
    }

    //asm volatile ("move.w   #0x2500, %sr"); // all interrupts off
    return byte;
}

//unsigned char input::kbgetchar()
uint32_t input::kbgetchar()
{
    /* When calling, make sure buffer is not empty!
     * check kbbuffcnt > 0 */
    //unsigned char byte=0;
    uint32_t byte=0;
    byte = *kboutpt; // Get byte
    kboutpt++; // Increment pointer

    if (kboutpt >= kbbuffer + BUFF_SIZE) // Pointer wrapping
    {
        kboutpt = kbbuffer;
    }

    kbbuffcnt--; // Decrement buffer count
    return byte;
}

unsigned char input::msgetchar()
{
    /* When calling, make sure buffer is not empty!
     * check msbuffcnt > 0 */
    unsigned char byte;
    byte = *msoutpt; // Get byte
    msoutpt++; // Increment pointer

    if (msoutpt >= msbuffer + BUFF_SIZE) // Pointer wrapping
    {
        msoutpt = msbuffer;
    }

    msbuffcnt--; // Decrement buffer count
    return byte;
}

uint16_t input::rs232getbuffcnt()
{
    return rs232buffcnt;
}

uint8_t input::kbgetbuffcnt()
{
    return kbbuffcnt;
}

uint8_t input::msgetbuffcnt()
{
    return msbuffcnt;
}

unsigned char input::mskbgetchar()
{
    /* When calling, make sure buffer is not empty!
     * check msbuffcnt > 0 */
    unsigned char byte;
    byte = *mskboutpt; // Get byte
    mskboutpt++; // Increment pointer

    if (mskboutpt >= mskbbuffer + BUFF_SIZE) // Pointer wrapping
    {
        mskboutpt = mskbbuffer;
    }

    mskbbuffcnt--; // Decrement buffer count
    return byte;
}

void input::dispatch()
{
    /* Reads from mskbbuff and dispatches to
     * kbbuff and msbuff */
    unsigned char byte;
    uint8_t packetpos;

    if (mskbbuffcnt > 0)
    {
        byte = mskbgetchar();

        if (byte == MAGIC_MOUSE) // MAGIC_MOUSE received => read 3-byte mouse packet
        {
            for (packetpos=0; packetpos<3; packetpos++)
            {
                while (mskbbuffcnt == 0);

                byte = mskbgetchar();

                if (byte != MAGIC_MOUSE)
                {
                    put_msbuff (byte);
                }
                else
                {
                    packetpos=0;
                }
            }
        }
        else
        {
            //put_kbbuff (byte);
            decode_scancode (byte);
        }
    }
}

void input::showMouse()
{
    v9990_SpritesEnable();
}

void input::hideMouse()
{
    v9990_SpritesDisable();
}

void input::MouseAction (screen *myScreen)
{
    uint8_t buffer[3];
    uint8_t position;
    static uint8_t lastLeftButton=0;

    if (msbuffcnt > 2) // Mouse packets have 3 bytes
    {
        for (position=0; position<3; position++)
        {
            buffer[position] = msgetchar();
//			if(buffer[position] == MAGIC_MOUSE)
//				position = 0;
        }

        if ((buffer[0]&0x08) && (buffer[0]&0xc0)==0) /* Bit 3 alway 1 */
        {
            if (buffer[0]&0x10) /* x sign bit set? */
            {
                x-= (uint8_t)~buffer[1]+1;
            }
            else
            {
                x+=buffer[1];
            }

            if (buffer[0]&0x20) /* y sign bit set? */
            {
                y+= (uint8_t)~buffer[2]+1;
            }
            else
            {
                y-=buffer[2];
            }

            if (x<0)
            {
                x=0;
            }

            if (x>511)
            {
                x=511;
            }

            if (y<0)
            {
                y=0;
            }

            if (y>211)
            {
                y=211;
            }

            low_x= (x&0xff);
            high_x= ((x>>8)&0x03)+64;
            low_y= (y&0xff);
            high_y= ((y>>8)&0x01);
            //v9990_SetBackdropColor(15-(uint8_t)(buffer[0]&0x07));
            vdp_critical |= 0x01;
            wait_vdp();
            v9990_SetVRAMWrite (VDP_CURSOR0_ATTRIB);
            asm volatile ("move.b   %0, 0x3df600"::"m" (low_y));	//
            asm volatile ("move.b   #0, 0x3df600"); /* pattern no 0 */
            asm volatile ("move.b   %0, 0x3df600"::"m" (high_y));	//
            asm volatile ("move.b   #0, 0x3df600"); /* pattern no 0 */
            asm volatile ("move.b   %0, 0x3df600"::"m" (low_x));	//
            asm volatile ("move.b   #0, 0x3df600"); /* pattern no 0 */
            asm volatile ("move.b   %0, 0x3df600"::"m" (high_x));	//
            asm volatile ("move.b   #0, 0x3df600"); /* pattern no 0 */
            vdp_critical &= ~0x01;

            if (buffer[0]&1) /* Left button */
            {
                if (lastLeftButton == 0) // only send signal once
                {
                    myScreen->leftButton (true, x, y);
                    lastLeftButton = 1;
                }

                // VDPWriteReg(VDP_LOP,VDP_LOP_WCSC);
                // v9990_SetCmdColor(0xffff);
                // box.left=x;
                // box.top=y;
                // box.width=4;
                // box.height=4;
                // v9990_DrawFilledBox(&box,0x0000);
                // //v9990_SetPoint(x,y);
            }
            else
                if (lastLeftButton == 1) // button was pressed => release
                {
                    myScreen->leftButton (false, x, y);
                    lastLeftButton = 0;
                }

            if (buffer[0]&2) /* Right button */
            {
                // VDPWriteReg(VDP_LOP,VDP_LOP_WCSC);
                // v9990_SetCmdColor(0xffff);
                // box.left=x;
                // box.top=y;
                // box.width=4;
                // box.height=4;
                // v9990_DrawFilledBox(&box,0x4444);
                // //v9990_SetPoint(x,y);
            }

            if (buffer[0]&4) /* Middle button */
            {
                // vdp_clearscreen();
            }
        }
    }
}

/* Keyboard routines */
uint32_t input::kbshift(uint16_t flag)
{
    uint16_t oldy;

    if (flag == -1)
        return (shifty);        /* return bitvector of shift state */

    oldy = shifty;
    shifty = flag;

    return (oldy);
}

/*
 * kbd_int : called by the interrupt routine for key events.
 */

void input::decode_scancode(uint8_t scancode)
{
    uint32_t value;                 /* the value to push into iorec */
    uint8_t ascii = 0;
    uint8_t scancode_only = scancode & ~KEY_RELEASED;  /* get rid of release bits */


    /* keyboard warm/cold reset */
//    if ((scancode == 0x53)  /* Del key and shifty is Alt+Ctrl but not LShift */
//        && ((shifty & (MODE_ALT|MODE_CTRL|MODE_LSHIFT)) == (MODE_ALT|MODE_CTRL))) {
//        if (shifty & MODE_RSHIFT) {
//            /* Ctrl+Alt+RShift+Del means cold reset */
//            cold_reset();
//        }
//        else {
//            /* Ctrl+Alt+Del means warm reset */
//            warm_reset();
//        }
//    }

    /* the additional mouse buttons use a separate vector */
    if (   scancode_only == 0x37  /* Mouse button 3 */
        || scancode_only == 0x5e  /* Mouse button 4 */
        || scancode_only == 0x5f  /* Mouse button 5 */
        || scancode_only == 0x59  /* Wheel up */
        || scancode_only == 0x5a  /* Wheel down */
        || scancode_only == 0x5c  /* Wheel left */
        || scancode_only == 0x5d) /* Wheel right */
    {
      //  mousexvec(scancode);
        return;
    }

    if (scancode & KEY_RELEASED) {
        /* stop key repeat */
        kb_ticks = 0;

        scancode = scancode_only;
        switch (scancode) {
        case KEY_RSHIFT:
            shifty &= ~MODE_RSHIFT;     /* clear bit */
            break;
        case KEY_LSHIFT:
            shifty &= ~MODE_LSHIFT;     /* clear bit */
            break;
        case KEY_CTRL:
            shifty &= ~MODE_CTRL;       /* clear bit */
            break;
        case KEY_ALT:
            shifty &= ~MODE_ALT;        /* clear bit */
            if(kb_altnum >= 0) {
                ascii = kb_altnum & 0xFF;
                kb_altnum = -1;
                goto push_value;
            }
            break;
        }
            return;
    }
    switch (scancode) {
    case KEY_RSHIFT:
        shifty |= MODE_RSHIFT;  /* set bit */
        return;
    case KEY_LSHIFT:
        shifty |= MODE_LSHIFT;  /* set bit */
        return;
    case KEY_CTRL:
        shifty |= MODE_CTRL;    /* set bit */
        return;
    case KEY_ALT:
        shifty |= MODE_ALT;     /* set bit */
        return;
    case KEY_CAPS:
        //if (conterm & 1) {
        //    keyclick(KEY_CAPS);
        //}
        shifty ^= MODE_CAPS;    /* toggle bit */
        return;
    }


    if (shifty & MODE_ALT) {
        const uint8_t *a;

        /* ALT-keypad means that char number */
        if (scancode >= 103 && scancode <= 112) {
            if (kb_altnum < 0) {
                kb_altnum = 0;
            }
            kb_altnum *= 10;
            kb_altnum += "\7\10\11\4\5\6\1\2\3\0" [scancode - 103];
            return;
        }

        if (shifty & MODE_SHIFT) {
            a = current_keytbl.altshft;
        } else if (shifty & MODE_CAPS) {
            a = current_keytbl.altcaps;
        } else {
            a = current_keytbl.altnorm;
        }
        while (*a && *a != scancode) {
            a += 2;
        }
        if (*a++) {
            ascii = *a;
        }
    } else if (shifty & MODE_SHIFT) {
        /* Fonction keys F1 to F10 */
        if (scancode >= 0x3B && scancode <= 0x44) {
            scancode += 0x19;
            goto push_value;
        }
        ascii = current_keytbl.shft[scancode];
    } else if (shifty & MODE_CAPS) {
        ascii = current_keytbl.caps[scancode];
    } else {
        ascii = current_keytbl.norm[scancode];
    }

    if (shifty & MODE_CTRL) {
        /* More complicated in TOS, but is it really necessary ? */
        ascii &= 0x1F;
    } else if(kb_dead >= 0) {
        //const uint8_t *a = current_keytbl.dead[kb_dead];
        //while (*a && *a != ascii) {
        //    a += 2;
        //}
        //if (*a++) {
        //    ascii = *a;
        //}
        kb_dead = -1;
    } else if(ascii <= DEADMAX && ascii >= DEADMIN) {
        kb_dead = ascii - DEADMIN;
        return;
    }

  push_value:
    value = ((uint32_t) scancode) << 16;
    value += ascii;
    //if (conterm & 0x8) {
    //    value += ((uint32_t) shifty) << 24;
    //}
#if DBG_KBD
    kprintf("KBD iorec: Pushing value 0x%08lx\n", value);
#endif
    //push_ikbdiorec(value);
    put_kbbuff(value);

    /* set initial delay for key repeat */
    kb_last_key = value;
    //kb_last_ikbdsys = kbdvecs.ikbdsys;
    kb_ticks = kb_initial;
}

//void input::push_ikbdiorec(uint32_t value)
//{
//}

/* Private parts */

void input::loadMousepointer()
{
    uint16_t i;
    char *pnt;
    vdp_critical |= 0x01;
    v9990_SetVRAMWrite (VDP_CURSOR0_PAT_DATA); // keep 1024*212

    for (i=0; i<sizeof (ms_arrow); i++)
    {
        ms_arrow[i]=~ms_arrow[i];
    }

    v9990_CopyRamToVram (ms_arrow,sizeof (ms_arrow));
    v9990_SetVRAMWrite (VDP_CURSOR0_ATTRIB);

    for (i=0; i< (2*4); i++)
    {
        asm volatile ("move.b   #0x10, 0x3df600"); // Clear all other sprites
        asm volatile ("move.b   #0x10, 0x3df600"); // clear all other sprites
    }

    WRITE_VDP (VDP_REGSEL,28);  // Sprites
    //WRITE_VDP(VDP_REGDAT,0x2);  // palette number
    WRITE_VDP (VDP_REGDAT,0x0); // palette number
    WRITE_VDP (VDP_REGSEL,25);  // Sprites base address
    WRITE_VDP (VDP_REGDAT,0x4); // 20000
    wait_vdp();
    i=0;
    /* set mousepointer attributes */
    v9990_SetVRAMWrite (VDP_CURSOR0_ATTRIB);
    asm volatile ("move.b   #50, 0x3df600");
    asm volatile ("move.b   #0, 0x3df600");
    asm volatile ("move.b   #0, 0x3df600");
    asm volatile ("move.b   #0, 0x3df600");
    asm volatile ("move.b   #50, 0x3df600");
    asm volatile ("move.b   #0, 0x3df600");
    asm volatile ("move.b   #64, 0x3df600");
    asm volatile ("move.b   #0, 0x3df600");
    showMouse();
    vdp_critical &= ~0x01;
}

// rs232

void input::setRS232()
{
    WRITEREG (DUART_CRA,0x10); /*  Reset MR1A pointer */
    WRITEREG (DUART_CRA,0x20); /*  Reset receiver */
    WRITEREG (DUART_CRA,0x30); /*  Reset transmitter */
    //WRITEREG(DUART_ACR,0x80); /*   */
    WRITEREG (DUART_ACR,0x00); /*   */
    WRITEREG (DUART_MR1A,0x93); /*  Set 8N1 */
    WRITEREG (DUART_MR2A,0x37); /*  Set normal mode */
    WRITEREG (DUART_CSRA,0xbb); /*  Set 9600 bps */
    WRITEREG (DUART_CRA,0x05); /*  Enable Rx and Tx */
    WRITEREG (DUART_OPRRST,0x05); /*  Set first RTS manually */
}
