/*
 * =====================================================================================
 *
 *       Filename:  interrupt.cpp
 *
 *    Description:  Interrupt class
 *
 *        Version:  1.0
 *        Created:  16.05.2016 09:13:29
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Simon Ferber (https://www.ist-schlau.de), simon@ist-schlau.de
 *   Organization:  Kiwi Homebrew Computer
 *
 * =====================================================================================
 */

#include "interrupt.h"
#include "input.h"
#include "screen.h"

screen *interrupt::root;
uint32_t interrupt::hz200;

void interrupt::exception()
{
    uint8_t color=0;

    while (1)
    {
        v9990_SetBackdropColor (color++);
    }
}

void interrupt::ISRtimer()
{
    if (! (hz200++%4)) // every 4th call
    {
        screen::timerISR (root);
    }

    asm volatile ("move.b   #1, 0x3dfb34"); /* clear IRQ */
}

void interrupt::ISRduart()
{
    //uint8_t ISR;
    //if((READREG(DUART_SRA)&1))
    //ISR =READREG(DUART_ISR);
    //asm volatile ("move.w   #0x2700, %sr"); // all interrupts off
    while (READREG (DUART_SRA) & 0x01)
    {
        input::put_rs232buff ((READREG (DUART_RBA))); // Get the byte from DUART and save to the input buffer
    }

    if ((READREG (DUART_SRB)&1))
        //while(READREG(DUART_ISR) & 0x20)
    {
        input::put_mskbbuff ((READREG (DUART_RBB))); // Get the byte from DUART and save to the input buffer
    }

    //asm volatile ("move.w   #0x2500, %sr"); // all interrupts off
}

/* Constructor switches all interrupts off
 * and initializes a default (just RTE)
 * routine for all IRQ */
interrupt::interrupt (screen *myScreen)
{
    asm volatile ("move.w   #0x2700, %sr"); // all interrupts off
    level = 7;
    (* ((long *)0x64) = (long)exception); // Interrupt level 1 (V9990)
    (* ((long *)0x68) = (long)exception); // Interrupt level 2 (IDE)
    (* ((long *)0x6c) = (long)exception); // Interrupt level 3 (CS8900a)
    (* ((long *)0x70) = (long)exception); // Interrupt level 4 (WD177x)
    (* ((long *)0x74) = (long)exception); // Interrupt level 5 (PI/T 2, timer)
    (* ((long *)0x78) = (long)exception); // Interrupt level 6 (DUART)
    (* ((long *)0x7c) = (long)exception); // Interrupt level 7
    /* Set exception vectors */
    (* ((long *)0x08) = (long)exception); // Bus Error
    (* ((long *)0x0c) = (long)exception); // Address Error
    (* ((long *)0x10) = (long)exception); // Illegal Instruction
    (* ((long *)0x14) = (long)exception); // Null Division
    (* ((long *)0x18) = (long)exception); // CHK Instruction
    (* ((long *)0x1c) = (long)exception); // Trapv Instruction
    (* ((long *)0x20) = (long)exception); // Privilege Violation
    (* ((long *)0x60) = (long)exception); // Spurious Interrupt
    root = myScreen;
}

interrupt::~interrupt()
{
}

void interrupt::setISR (void (*ISR)(), uint8_t lvl)
{
    switch (lvl)
    {
    case 1:
        (* ((long *)0x64) = (long)ISR); // Interrupt level 1 (V9990)
        break;

    case 2:
        (* ((long *)0x68) = (long)ISR); // Interrupt level 2 (IDE)
        break;

    case 3:
        (* ((long *)0x6c) = (long)ISR); // Interrupt level 3 (CS8900a)
        break;

    case 4:
        (* ((long *)0x70) = (long)ISR); // Interrupt level 4 (WD177x)
        break;

    case 5:
        (* ((long *)0x74) = (long)ISR); // Interrupt level 5 (PI/T 2, timer)
        break;

    case 6:
        (* ((long *)0x78) = (long)ISR); // Interrupt level 6 (DUART)
        break;

    case 7:
        (* ((long *)0x7c) = (long)ISR); // Interrupt level 7
        break;
    }
}

void interrupt::setLevel (uint8_t lvl)
{
    switch (lvl)
    {
    case 1:
        asm volatile ("move.w   #0x2100, %sr");
        break;

    case 2:
        asm volatile ("move.w   #0x2200, %sr");
        break;

    case 3:
        asm volatile ("move.w   #0x2300, %sr");
        break;

    case 4:
        asm volatile ("move.w   #0x2400, %sr");
        break;

    case 5:
        asm volatile ("move.w   #0x2500, %sr");
        break;

    case 6:
        asm volatile ("move.w   #0x2600, %sr");
        break;

    case 7:
        asm volatile ("move.w   #0x2700, %sr");
        break;
    }

    level = lvl;
}

void interrupt::initTimer()
{
    WRITE_PITB (TCR,0xe0); // stop timer
    WRITE_PITB (CPRH,0x00); // preload timer (high,medium,low) 10000
    WRITE_PITB (CPRM,0x06);
    WRITE_PITB (CPRL,0x1b);
    WRITE_PITB (TCR,0xe1); // start timer
    hz200 = 0;
}

/* Private */
