/*
 * =====================================================================================
 *
 *       Filename:  interrupt.h
 *
 *    Description:  Interrupt class
 *
 *        Version:  1.0
 *        Created:  12.05.2016 21:26:23
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Simon Ferber (https://www.ist-schlau.de), simon@ist-schlau.de
 *   Organization:  Kiwi Homebrew Computer
 *
 * =====================================================================================
 */

#ifndef INTERRUPT_H
#define INTERRUPT_H

#include <cstdint>
#include "screen.h"

#define PITB_BASE        0x003dfb00
#define PSR             0x1a
#define TCR             0x20
#define CPRH            0x26
#define CPRM            0x28
#define CPRL            0x2A
#define CRH             0x2e
#define CRM             0x30
#define CRL             0x32
#define TSR             0x34
#define READ_PITB(x)    (*((volatile char *) PITB_BASE + (x) ))
#define WRITE_PITB(x, y)        (*((char *) PITB_BASE + (x) ) = (y) )

class interrupt
{
public:
    interrupt (screen *myScreen); // Constructor
    ~interrupt(); // Destructor
    void setISR (void (*ISR)(), uint8_t level);
    void setLevel (uint8_t level);
    void initTimer();
    static void ISRtimer() __attribute__ ((interrupt));
    static void ISRduart() __attribute__ ((interrupt));
    static void exception() __attribute__ ((interrupt));
    static screen *root;
    static uint32_t hz200;
private:
    uint8_t level;
};

#endif
