/*
 * =====================================================================================
 *
 *       Filename:  screen.c
 *
 *    Description:  Screen class
 *
 *        Version:  1.0
 *        Created:  15.05.2016 09:23:45
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Simon Ferber (https://www.ist-schlau.de), simon@ist-schlau.de
 *   Organization:  Kiwi Homebrew Computer
 *
 * =====================================================================================
 */

#include "screen.h"
#include "textbox.h"
#include "widget.h"
#include "topaz.h"


/* Public parts */
volatile uint8_t vdp_critical;

/* Constructor */
screen::screen()
{
//	return;
    bgColor=0x4444;
    screenbox.left = 0;
    screenbox.top= 0;
    screenbox.width= 512;
    screenbox.height= 212;
    v9990_SpritesDisable();
    v9990_SetScreenMode (VDP_MODE_B3, VDP_SCR0_4BIT, VDP_SCR0_XIM1024, VDP_SCR1_PAL, VDP_PAL_CTRL_PAL);
    //v9990_SetScreenMode (VDP_MODE_B3, VDP_SCR0_4BIT, VDP_SCR0_XIM1024, VDP_SCR1_PAL + VDP_SCR1_EO + VDP_SCR1_IL, VDP_PAL_CTRL_PAL);
    //v9990_SetScreenMode (VDP_MODE_B4, VDP_SCR0_4BIT, VDP_SCR0_XIM1024, VDP_SCR1_PAL, VDP_PAL_CTRL_PAL);
    v9990_SetScrollX (0);
    v9990_SetScrollY (0);
    activeWidget = nullptr;
    loadFont();
    clearScreen();
    vdp_critical = 0;
}

/* Destructor */
screen::~screen()
{
}

/* timer interrupt service */
void screen::timerISR (screen *myScreen)
{
    std::vector<widget*> *myWidgets = &myScreen->myWidgets;

    for (uint8_t i=0; i < myWidgets->size(); i++)
    {
        if ( ((*myWidgets)[i]->getType() == widget::type::Textbox) || ((*myWidgets)[i]->getType() == widget::type::Textinput) )
        {
            textbox *t = (textbox *) (*myWidgets)[i];
            t->blink();
        }
    }
}

void screen::clearScreen()
{
    vdp_critical |= 0x01;
    v9990_SetCmdWriteMask (0xffff);
    VDPWriteReg (VDP_LOP,VDP_LOP_WCSC);
    v9990_DrawFilledBox (&screenbox, bgColor);
    wait_vdp();
    vdp_critical &= ~0x01;
}

void screen::addWidget (widget *newWidget)
{
    //widget *w = new widget();
    newWidget->topScreen = this;
    newWidget->setWidgetNumber(myWidgets.size());
    //newWidget->setPreviousWidget();
    myWidgets.push_back (newWidget);
}

void screen::printWidgets()
{
    for (uint8_t i=0; i < myWidgets.size(); i++)
    {
        //std::cout << myWidgets[i]->getX() << " - " << myWidgets[i]->getY() << std::endl;
        myWidgets[i]->render();
    }
}

void screen::leftButton (bool pressed, uint16_t x, uint16_t y)
{
    uint16_t widgetX, widgetX2, widgetY, widgetY2;

    for (uint8_t i=0; i < myWidgets.size(); i++)
    {
        widgetX = myWidgets[i]->getX();
        widgetY = myWidgets[i]->getY();
        widgetX2 = widgetX + myWidgets[i]->getWidth();
        widgetY2 = widgetY + myWidgets[i]->getHeight();

        if (myWidgets[i]->getlastAction() == widget::action::Selected && !pressed)
        {
            myWidgets[i]->doAction (widget::action::None);
        }

        if (widgetX < x && widgetX2 > x)
        {
            if (widgetY < y && widgetY2 > y)
            {
                if (pressed)
                {
                    myWidgets[i]->doAction (widget::action::Selected);

                    if (myWidgets[i]->callback != nullptr)
                    {
                        //myWidgets[i]->callback (this->myWidgets[4]);
                        myWidgets[i]->callback (myWidgets[i]); /* Tell the callback func who is calling */
                    }
                }
                else
                {
                    myWidgets[i]->doAction (widget::action::None);
                }
            }
        }
    }
}

widget *screen::getActiveWidget()
{
    return activeWidget;
}

void screen::setActiveWidget (widget *w)
{
    activeWidget = w;
}

widget *screen::getWidget(uint8_t number)
{
    return myWidgets[number];
}

std::vector<widget*> screen::getMyWidgets()
{
    return myWidgets;
}

/* Private parts */
void screen::loadFont()
{
    vdp_critical |= 0x01;
    v9990_SetVRAMWrite (VRAM_FONT); // keep 1024*212
    v9990_CopyRamToVram (fnt_topaz,sizeof (fnt_topaz));
    vdp_critical &= ~0x01;
}
