/*
 * =====================================================================================
 *
 *       Filename:  screen.h
 *
 *    Description:  Screen class
 *
 *        Version:  1.0
 *        Created:  12.05.2016 21:25:17
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Simon Ferber (https://www.ist-schlau.de), simon@ist-schlau.de
 *   Organization:  Kiwi Homebrew Computer
 *
 * =====================================================================================
 */

#ifndef SCREEN_H
#define SCREEN_H

#define VRAM_FONT 0x20000

#include "widget.h"
#include "hammacro.h"

#include <vector>
#include <cstdint>
#include <cstdbool>

extern "C" {
#include <v9990.h>
}

class textbox;
class textinput;

extern volatile uint8_t vdp_critical;

class screen
{
public:
    screen(); // Constructor
    ~screen(); // Destructor
    //void setScreenMode();
    static void timerISR (screen *myScreen);
    void clearScreen();
    void addWidget (widget *newWidget);
    std::vector<widget*> getMyWidgets();

    void printWidgets();
    void leftButton (bool pressed, uint16_t x, uint16_t y);
    widget *getActiveWidget();
    widget *getWidget(uint8_t number);
    void setActiveWidget (widget *);
    macro *macros;
private:
    VDP_BOX screenbox;
    std::vector<widget*> myWidgets;
    widget *activeWidget;
    uint16_t bgColor;
    void loadFont();
};

#endif
