/*
 * =====================================================================================
 *
 *       Filename:  textbox.cpp
 *
 *    Description:  Textbox class
 *
 *        Version:  1.0
 *        Created:  15.05.2016 16:59:38
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Simon Ferber (https://www.ist-schlau.de), simon@ist-schlau.de
 *   Organization:  Kiwi Homebrew Computer
 *
 * =====================================================================================
 */

#include "textbox.h"

textbox::textbox (uint16_t x, uint16_t y, uint16_t w, uint16_t h)
{
    positionX = x;
    positionY = y;
    width = w;
    height = h;
    widgetType = type::Textbox;
    callback = nullptr;
    tBlink = false;
    /* vt52 init */
    v_cel_ht = 8;
    v_cel_wr = (w>>3);
    v_cel_mx = (w>>3)-1;
    v_cel_my = (h>>3)-1;
    tpositionX = positionX + 2;
    tpositionY = positionY + 2;
    twidth = v_cel_mx << 3; // width - 4;
    theight = v_cel_my << 3; // height - 4;
    /* Initial cursor settings */
    v_cur_cx = 0;                       // cursor to column 0
    v_cur_cy = 0;                       // cursor to line 0
    v_cur_of = 0;                       // line offset is 0
    cursconf (4, 30);                   // .5 second blink rate (@60 Hz vblank).
    cursconf (3, 0);                   // cursor on
    v_cur_tim = v_period;               // load initial value to blink timer
    disab_cnt = 0;                      // cursor disabled 1 level deep.
    v_col_fg = 0x0000;
    v_col_bg = 0x5555;
    con_state = &textbox::normal_ascii;           /* Init conout state machine */
    clear_and_home();
    setActive (false);
    render();
    v_stat_0 = M_CFLASH;
}

textbox::~textbox()
{
}

void textbox::render()
{
    //std::cout << "Ich bin eine Textbox bei x=" << getX() << ", y=" << getY() << std::endl;
    uint16_t color1, color2;
    color1 = 0x6666;
    color2 = 0x6666;
    //color2 = 0x3333;
    static uint16_t color3=0x5555;
    // Draw button background
    VDP_BOX btn_background, caption_box;
    uint16_t btn_bgcolor=color3;//0x2525;
    //color3 += 0x1111;
    btn_background.left = positionX;
    btn_background.top = positionY;
    btn_background.width = width;
    btn_background.height = height;
    vdp_critical |= 0x01;
    v9990_SetCmdWriteMask (0xffff);
    VDPWriteReg (VDP_LOP,VDP_LOP_WCSC);
    v9990_DrawFilledBox (&btn_background, btn_bgcolor++);
    wait_vdp();
    // Draw button border
    v9990_DrawLine (positionX, positionY, width, 0, color1); // top border
    v9990_DrawLine (positionX, positionY, 0, height, color1); // left border
    v9990_DrawLine (positionX, positionY + height, width, 0, color2); // bottom border
    v9990_DrawLine (positionX + width, positionY, 0, height, color2); // right border
    vdp_critical &= ~0x01;
}

/* Editor */

void textbox::edit(uint32_t key)
{
    uint32_t scancode;
    scancode = (key >> 16) & 0xff;

    if(scancode == 0x0f) /* TAB */
    {
        
    }
    cputc(key);
}

/* VT52 */

void textbox::cputc (char c)
{
    if (!con_state)
    {
        /* vt52_init() has not been called yet, ignore */
        return;
    }

    /* based on our state goto the correct stub routine */
    (this->*con_state) (c & 0xff);
}

void textbox::ascii_out (int ch)
{
    bool visible;                       /* was the cursor visible? */
    visible = v_stat_0 & M_CVIS;        /* test visibility bit */

    if (visible)
    {
        neg_cell (v_cur_cx, v_cur_cy);            /* delete cursor. */
        v_stat_0 &= ~M_CVIS;                    /* start of critical section */
    }

    cell_xfer (ch , v_cur_cx, v_cur_cy);

    if (next_cell())
    {
        uint8_t * cell;
        int y = v_cur_cy;
        /* perform cell carriage return. */
        //cell = v_bas_ad + (ULONG)v_cel_wr * y;
        v_cur_cx = 0;                   /* set X to first cell in line */

        /* perform cell line feed. */
        if (y < v_cel_my)
        {
            cell += v_cel_wr;           /* move down one cell */
            v_cur_cy = y + 1;           /* update cursor's y coordinate */
        }
        else
        {
            scroll_up (0);              /* scroll from top of screen */
        }

        //v_cur_ad = cell;                /* update cursor address */
    }

    /* if visible */
    if (visible)
    {
        //neg_cell(v_cur_ad);             /* display cursor. */
        neg_cell (v_cur_cx, v_cur_cy);            /* display cursor. */
        v_stat_0 |= M_CSTATE;           /* set state flag (cursor on). */
        v_stat_0 |= M_CVIS;             /* end of critical section. */

        /* do not flash the cursor when it moves */
        if (v_stat_0 & M_CFLASH)
        {
            v_cur_tim = v_period;       /* reset the timer. */
        }
    }
}

void textbox::normal_ascii (uint16_t ch)
{
    /* If the character is printable ascii */
    if (ch >= 0x20)
    {
        ascii_out (ch);   /* go print it. */
    }
    /* We handle the following control characters as special: */
    /* 7 = bell */
    /* 8 = backspace */
    /* 9 = Horizontal tab */
    /* 10 = Line feed */
    /* 11 = Vertical tab (Treated as line feed) */
    /* 12 = Form Feed (Treated as line feed) */
    /* 13 = Carriage Return */
    /* 27 = Escape (Start Command) */
    /* If escape character alter next state */
    else
        if (ch == 0x1b)
        {
            /* handle the control characters */
            con_state = &textbox::esc_ch1;    /* set constate to handle esc codes */
        }
        /* Other control haracters */
        else
            if (ch >= 7 && ch <= 13)
            {
                (this->*cntl_tab[ch - 7])();
            }

    /* All others are thrown away. */
}

void textbox::cell_xfer (uint8_t character, uint8_t x, uint8_t y)
{
    VDP_BOX char_bbox;
    uint16_t fg;
    uint16_t bg;

    /* check for reversed foreground  and background colors */
    if (v_stat_0 & M_REVID)
    {
        fg = v_col_bg;
        bg = v_col_fg;
    }
    else
    {
        fg = v_col_fg;
        bg = v_col_bg;
    }

//    fg=0x66;
//    bg=0x44;
    fg= (fg&0x0f)+ ((fg<<4)&0xf0);
    bg= (bg&0x0f)+ ((bg<<4)&0xf0);
    char_bbox.left=tpositionX + (x<<3);
    char_bbox.top=tpositionY + (y<<3);
    char_bbox.width = 8;
    char_bbox.height = 8;
    wait_vdp();
    VDPWriteReg (VDP_ARG,0);
    VDPWriteReg (VDP_LOP,VDP_LOP_WCSC);
    VDPWriteReg (VDP_FC,fg); // Font color
    WRITE_VDP (VDP_REGDAT,fg);
    VDPWriteReg (VDP_BC,bg); // Backdop color
    WRITE_VDP (VDP_REGDAT,bg);
    vdp_critical |= 0x01;
    v9990_SetCmdWriteMask (0xffff);
    character-=32;
    v9990_CopyVRamCharToXY (&char_bbox, VRAM_FONT + (character<<3));
    vdp_critical &= ~0x01;
}

void textbox::neg_cell (uint8_t x, uint8_t y)
{
    VDP_BOX box;
    v_stat_0 |= M_CRIT;                 /* start of critical section. */
    box.left = tpositionX + (x<<3);
    box.top = tpositionY + (y<<3);
    box.width= 8;
    box.height= 8;
    wait_vdp();
    vdp_critical |= 0x01;
    v9990_SetCmdWriteMask (0xffff);
    VDPWriteReg (VDP_LOP,VDP_LOP_WCEORSC);
    v9990_DrawFilledBox (&box, 0xffff);
    wait_vdp();
    v_stat_0 &= ~M_CRIT;                /* end of critical section. */
    VDPWriteReg (VDP_LOP,VDP_LOP_WCSC);
    vdp_critical &= ~0x01;
}

void textbox::invert_cell (uint16_t celx, uint16_t cely)
{
    neg_cell (celx, cely);
}

bool textbox::next_cell()
{
    /* check bounds against screen limits */
    if (v_cur_cx == v_cel_mx)                   /* increment cell ptr */
    {
        if (! (v_stat_0 & M_CEOL))
        {
            /* overwrite in effect */
            return 0;                   /* no wrap condition exists */
            /* dont change cell parameters */
        }

        /* call carriage return routine */
        /* call line feed routine */
        return 1;                       /* indicate that CR LF is required */
    }

    v_cur_cx += 1;                      /* next cell to right */

    /* if X is even, move to next word in the plane */
    if (v_cur_cx & 1)
    {
        /* x is odd */
        //v_cur_ad += 1;                  /* a1 -> new cell */
        return 0;                       /* indicate no wrap needed */
    }

    /* new cell (1st plane), added offset to next word in plane */
    //v_cur_ad += (v_planes << 1) - 1;
    return 0;                           /* indicate no wrap needed */
}

void textbox::move_cursor (int x, int y)
{
    int old_cx, old_cy;
    /* update cell position */

    /* clip x. */
    if (x > v_cel_mx)        /* no, branch. */
    {
        x = v_cel_mx;       /* yes. */
    }

    /* clip y. */
    if (y > v_cel_my)        /* no, branch. */
    {
        y = v_cel_my;       /* yes. */
    }

    old_cx = v_cur_cx;
    old_cy = v_cur_cy;
    v_cur_cx = x;
    v_cur_cy = y;

    /* is cursor visible? */
    if (! (v_stat_0 & M_CVIS))
    {
        /* not visible */
        //v_cur_ad = cell_addr(x, y);             /* just set new coordinates */
        return;                                 /* and quit */
    }

    /* is cursor flashing? */
    if (v_stat_0 & M_CFLASH)
    {
        v_stat_0 &= ~M_CVIS;                    /* yes, make invisible...semaphore. */

        /* is cursor presently displayed ? */
        if (! (v_stat_0 & M_CSTATE))
        {
            /* not displayed */
            //v_cur_ad = cell_addr(x, y);         /* just set new coordinates */
            /* show the cursor when it moves */
            //neg_cell(v_cur_ad);                 /* complement cursor. */
            neg_cell (x, y);                /* complement cursor. */
            v_stat_0 |= M_CSTATE;
            v_cur_tim = v_period;               /* reset the timer. */
            v_stat_0 |= M_CVIS;                 /* end of critical section. */
            return;
        }
    }

    /* move the cursor after all special checks failed */
    //neg_cell(v_cur_ad);                         /* erase present cursor */
    neg_cell (old_cx, old_cy);                        /* erase present cursor */
    //v_cur_ad = cell_addr(x, y);                 /* fetch x and y coords. */
    //neg_cell(v_cur_ad);                         /* complement cursor. */
    neg_cell (x, y);                        /* complement cursor. */
    /* do not flash the cursor when it moves */
    v_cur_tim = v_period;                       /* reset the timer. */
    v_stat_0 |= M_CVIS;                         /* end of critical section. */
}

void textbox::cursor_on_cnt (void)
{
    /* if disable count is zero (cursor still shown) then return */
    if (!disab_cnt)
    {
        return;
    }

    disab_cnt--;                        /* decrement the disable counter. */

    if (!disab_cnt)
    {
        do_cnt_esce();    /* if 0, do the enable. */
    }
}

void textbox::cursor_on()
{
    /* if disable count is zero (cursor still shown) then return */
    if (!disab_cnt)
    {
        return;
    }

    disab_cnt = 0;                      /* reset the disable counter. */
    do_cnt_esce();
}

void textbox::cursor_off()
{
    disab_cnt++;

    /* test and clear the visible state bit. */
    if (! (v_stat_0 & M_CVIS))
    {
        return;    /* if already invisible, just return */
    }

    v_stat_0 &= ~M_CVIS;                /* make invisible! */

    /* see, if flashing is disabled */
    if (! (v_stat_0 & M_CFLASH))
    {
        invert_cell (v_cur_cx, v_cur_cy);
    }
    /* see, if cursor is on or off */
    else
        if (v_stat_0 & M_CSTATE)
        {
            v_stat_0 &= ~M_CSTATE;    /* cursor off? */
            invert_cell (v_cur_cx, v_cur_cy);
        }
}

void textbox::do_cnt_esce()
{
    invert_cell (v_cur_cx, v_cur_cy);       /* complement cursor. */
    v_stat_0 |= M_CVIS;                     /* set visibility bit. */

    /* see if flashing is enabled. */
    if (v_stat_0 & M_CFLASH)
    {
        v_stat_0 |= M_CSTATE;                   /* set cursor on. */
        /* do not flash the cursor when it moves */
        v_cur_tim = v_period;                   /* reset the timer. */
    }
}

void textbox::set_fg (void)
{
    con_state = &textbox::get_fg_col;             /* Next char is the FG color */
}

void textbox::set_bg (void)
{
    con_state = &textbox::get_bg_col;             /* Next char is the BG color */
}

void textbox::clear_and_home()
{
    cursor_off();                               /* hide cursor. */
    move_cursor (0, 0);                         /* cursor home */
    blank_out (0, 0, v_cel_mx, v_cel_my);       /* clear screen. */
    cursor_on_cnt();                            /* show cursor. */
}

void textbox::esc_ch1 (uint16_t ch)
{
    con_state = &textbox::normal_ascii;   /* Most functions only 2 chars so set */
    ch -= 0x41;                 /* state to normal ascii.  Bias by low */

    if (ch < 0)
    {
        return;    /* char and get out if invalid */
    }

    /* If in the range A-M go handle */
    if (ch <= 12)
    {
        (this->*am_tab[ch])();
    }
    /* check for lower case control code */
    else
        if (ch != 24)
        {
            ch -= 0x21;       /* see if b to w */

            if (ch >= 0 && ch <= 21)
            {
                (this->*bw_tab[ch])();
            }
        }
        else
            /* direct cursor addressing, take 2 additional chars */
        {
            con_state = &textbox::get_row;
        }
}

void textbox::get_row (uint16_t ch)
{
    save_row = ch - 0x20;       /* Remove space bias */
    con_state = &textbox::get_column;
}

void textbox::get_fg_col (uint16_t ch)
{
    /* set the foreground color, low-order bits only. */
    v_col_fg = (ch - 0x20) & 0x0f;
    con_state = &textbox::normal_ascii;           /* Next char is not special */
}

void textbox::get_bg_col (uint16_t ch)
{
    /* set the foreground color, low-order bits only. */
    v_col_bg = (ch - 0x20) & 0x0f;
    con_state = &textbox::normal_ascii;           /* Next char is not special */
}

void textbox::get_column (uint16_t ch)
{
    int row, col;
    col = ch - 0x20;                    /* Remove space bias */
    row = save_row;
    move_cursor (col,row);
    con_state = &textbox::normal_ascii;           /* Next char is not special */
}

void textbox::blank_out (int topx, int topy, int botx, int boty)
{
    uint16_t color = v_col_bg;             /* bg color value */
    //color=0x0000;
    VDP_BOX box;
    box.left = tpositionX + (topx<<3);
    box.top = tpositionY + (topy<<3);
    box.width= 8+ ((botx-topx)<<3);
    box.height= 8+ ((boty-topy)<<3);
    wait_vdp();
    vdp_critical |= 0x01;
    v9990_SetCmdWriteMask (0xffff);
    VDPWriteReg (VDP_LOP,VDP_LOP_WCSC);
    v9990_DrawFilledBox (&box, color);
    vdp_critical &= ~0x01;
}

uint16_t textbox::cursconf (uint16_t function, uint16_t operand)
{
    switch (function)
    {
    case 0:
        cursor_off();                   /* set cursor unvisible */
        break;

    case 1:
        cursor_on();                    /* set cursor visible */
        break;

    case 2:
        v_stat_0 &= ~M_CFLASH;          /* unset cursor flash bit */
        break;

    case 3:
        v_stat_0 |= M_CFLASH;           /* set cursor flash bit */
        break;

    case 4:
        v_period = operand;             /* set cursor flash interval */
        break;

    case 5:
        return (v_period);              /* set cursor flash interval */
    }

    return 0;
}


void textbox::scroll_up (int top_line)
{
    VDP_COPY_XY_XY box;
    box.sourceX=tpositionX;
    box.sourceY=tpositionY + ((top_line+1)<<3);
    box.destX=tpositionX;
    box.destY=tpositionY + (top_line<<3);
    box.width=twidth;
    box.height=theight- (top_line<<3);
    vdp_critical |= 0x01;
    if(box.height>=8) /* correct workaround for 1-line box? */
    {
        wait_vdp();
        v9990_CopyXYToXY (&box);
        wait_vdp();
    }
    /* exit thru blank out, bottom line cell address y to top/left cell */
    blank_out (0, v_cel_my , v_cel_mx, v_cel_my);
    wait_vdp();
    vdp_critical &= ~0x01;
}

void textbox::delete_line (void)
{
    cursor_off();               /* hide cursor. */
    scroll_up (v_cur_cy);       /* scroll up 1 line & blank bottom line. */
    move_cursor (0, v_cur_cy);  /* move cursor to beginning of line. */
    cursor_on_cnt();            /* show cursor. */
}

void textbox::erase_from_home (void)
{
    erase_from_bol_impl(); /* erase from beginning of line. */

    /* first line? */
    if (!v_cur_cy)
    {
        return;    /* yes, done. */
    }

    /* erase rest of screen. */
    blank_out (0, 0, v_cel_mx, v_cur_cy - 1);        /* clear screen. */
}

void textbox::line_wrap_on (void)
{
    v_stat_0 |= M_CEOL;    /* set the eol handling bit. */
}

void textbox::line_wrap_off (void)
{
    v_stat_0 &= ~M_CEOL;    /* clear the eol handling bit. */
}

void textbox::ascii_cr (void)
{
    /* beginning of current line. */
    move_cursor (0, v_cur_cy);
}

void textbox::ascii_lf (void)
{
    /* at bottom of screen? */
    if (v_cur_cy != v_cel_my)
    {
        cursor_down_impl();
    }
    else
    {
        cursor_off();                   /* yes, hide cursor. */
        scroll_up (0);                  /* scroll up 1 line */
        cursor_on_cnt();                /* show cursor. */
    }
}

void textbox::cursor_down_impl (void)
{
    if (v_cur_cy != v_cel_my)
    {
        move_cursor (v_cur_cx, v_cur_cy + 1);
    }
}

void textbox::nop()
{
    return;
}

void textbox::do_bell (void)
{
//    if (conterm & 4) {
//        bell();
//    }
}

void textbox::do_backspace (void)
{
    cursor_left_impl();
    {
    /*  Test delete character */
        bool visible;                       /* was the cursor visible? */
        visible = v_stat_0 & M_CVIS;        /* test visibility bit */

        if (visible)
        {
            neg_cell (v_cur_cx, v_cur_cy);            /* delete cursor. */
            v_stat_0 &= ~M_CVIS;                    /* start of critical section */
        }
        cell_xfer (' ' , v_cur_cx, v_cur_cy);
        if (visible)
        {
            //neg_cell(v_cur_ad);             /* display cursor. */
            neg_cell (v_cur_cx, v_cur_cy);            /* display cursor. */
            v_stat_0 |= M_CSTATE;           /* set state flag (cursor on). */
            v_stat_0 |= M_CVIS;             /* end of critical section. */

            /* do not flash the cursor when it moves */
            if (v_stat_0 & M_CFLASH)
            {
                v_cur_tim = v_period;       /* reset the timer. */
            }
        }
    /*  Test delete character */
    }
}

void textbox::do_tab (void)
{
    move_cursor ((v_cur_cx & 0xfff8) + 8, v_cur_cy);
}

void textbox::cursor_down (void)
{
    cursor_down_impl();
}

void textbox::cursor_up (void)
{
    if (v_cur_cy)
    {
        move_cursor (v_cur_cx, v_cur_cy - 1);
    }
}

void textbox::cursor_right (void)
{
    if (v_cur_cx != v_cel_mx)
    {
        move_cursor (v_cur_cx + 1, v_cur_cy);
    }
}

void textbox::cursor_left_impl (void)
{
    if (v_cur_cx)
    {
        move_cursor (v_cur_cx - 1, v_cur_cy);
    }
}

void textbox::cursor_left (void)
{
    cursor_left_impl();
}

void textbox::cursor_home (void)
{
    move_cursor (0, 0);
}

void textbox::reverse_linefeed (void)
{
    /* if not at top of screen */
    if (v_cur_cy)
    {
        move_cursor (v_cur_cx, v_cur_cy - 1);
    }
    else
    {
        int savex = v_cur_cx;           /* save current x position */
        insert_line();                  /* Insert a line */
        move_cursor (savex, 0);
    }
}

void textbox::erase_to_eos (void)
{
    erase_to_eol_impl(); /* erase to end of line. */

    /* last line? */
    if (v_cur_cy == v_cel_my)
    {
        return;    /* yes, done. */
    }

    /* erase from upper left corner to lower right corner. */
    blank_out (0, v_cur_cy + 1, v_cel_mx, v_cel_my);
}

void textbox::erase_to_eol_impl (void)
{
    bool wrap = v_stat_0 & M_CEOL;      /* save line wrap status */
    v_stat_0 &= ~M_CEOL;    /* clear EOL handling bit. (overwrite) */
    cursor_off();               /* hide cursor. */
    save_cursor_pos();          /* save cursor position. */

    /* is x = x maximum? */
    if (v_cur_cx == v_cel_mx)
    {
        ascii_out (0x20);    /* output a space, the cell is odd!. */
    }
    else
    {
        /* test, if x is even or odd */
        if (v_cur_cx & 0x1)
        {
            ascii_out (0x20);    /* first output a space. */
        }

        blank_out (v_cur_cx, v_cur_cy, v_cel_mx, v_cur_cy);
    }

    /* restore wrap flag, the result of EOL test. */
    if (wrap)
    {
        v_stat_0 |= M_CEOL;
    }

    restore_cursor_pos();       /* restore cursor position. */
    cursor_on_cnt();            /* show cursor. */
}

void textbox::erase_to_eol (void)
{
    erase_to_eol_impl();
}

void textbox::reverse_video_on (void)
{
    v_stat_0 |= M_REVID;    /* set the reverse bit. */
}

void textbox::reverse_video_off (void)
{
    v_stat_0 &= ~M_REVID;    /* clear the reverse bit. */
}

void textbox::insert_line (void)
{
    cursor_off();               /* hide cursor. */
    scroll_down (v_cur_cy);     /* scroll down 1 line & blank current line. */
    move_cursor (0, v_cur_cy);  /* move cursor to beginning of line. */
    cursor_on_cnt();            /* show cursor. */
}

void textbox::save_cursor_pos (void)
{
    v_stat_0 |= M_SVPOS;    /* set "position saved" status bit. */
    /* save the x and y coords of cursor. */
    sav_cur_x = v_cur_cx;
    sav_cur_y = v_cur_cy;
}

void textbox::restore_cursor_pos (void)
{
    if (v_stat_0 & M_SVPOS)
    {
        move_cursor (sav_cur_x, sav_cur_y);    /* move to saved position. */
    }
    else
    {
        move_cursor (0, 0);    /* if position was not saved, home cursor. */
    }

    v_stat_0 &= ~M_SVPOS;    /* clear "position saved" status bit. */
}

void textbox::erase_line (void)
{
    cursor_off();               /* hide cursor. */
    blank_out (0, v_cur_cy, v_cel_mx, v_cur_cy);   /* blank whole line. */
    move_cursor (0, v_cur_cy);  /* move cursor to beginning of line. */
    cursor_on_cnt();            /* show cursor. */
}

void textbox::erase_from_bol_impl (void)
{
    cursor_off();               /* hide cursor. */
    save_cursor_pos();          /* save cursor position. */

    /* are we in column 0?*/
    if (v_cur_cx == 0)
    {
        ascii_out (0x20);    /* output a space. */
    }
    else
    {
        /* test, if x is even or odd */
        if (v_cur_cx & 1)
        {
            ascii_out (0x20);   /* first output a space. */
            blank_out (0, v_cur_cy, v_cur_cx - 2, v_cur_cy);
        }
        else
        {
            blank_out (0, v_cur_cy, v_cur_cx, v_cur_cy);
        }
    }

    restore_cursor_pos();       /* restore cursor position. */
    cursor_on_cnt();            /* show cursor. */
}

void textbox::erase_from_bol (void)
{
    erase_from_bol_impl();
}

void textbox::scroll_down (int start_line)
{
    VDP_COPY_XY_XY box;
    box.sourceX=tpositionX+twidth;
    box.sourceY=tpositionY + theight-1;
    box.destX=tpositionX+twidth;
    box.destY=tpositionY + theight + 7;
    box.width=twidth;
    box.height=theight- (start_line<<3);
    vdp_critical |= 0x01;
    if(box.height>=8) /*  scroll if not in last line */
    {
        wait_vdp();
        VDPWriteReg (VDP_ARG,0x0c); /* DIX=1, DIY=1 */
        v9990_CopyXYToXY (&box);
        wait_vdp();
        VDPWriteReg (VDP_ARG,0);
    }
    /* exit thru blank out */
    blank_out(0, start_line , v_cel_mx, start_line );
    wait_vdp();
    vdp_critical &= ~0x01;
}

void textbox::blink (void)
{
    static uint8_t i=0;

//    if(isActive() == true && tBlink == true) {
//    if(tBlink == true) {

    /* test critical VDP state */
    if (vdp_critical)
    {
        return;
    }

    /* test visibility/semaphore bit. */
    if (! (v_stat_0 & M_CVIS))
    {
        return;    /* if invisible or blocked, return. */
    }

    /* test flash bit. */
    if (! (v_stat_0 & M_CFLASH))
    {
        return;    /* if not flashing, return. */
    }

    //			v9990_SetBackdropColor(i++);
    /* decrement cursor flash timer. */
    if (--v_cur_tim)
    {
        return;    /* if <> 0, return. */
    }

    v_cur_tim = v_period;       /* else reset timer. */

    /* toggle cursor state. */
    if (v_stat_0 & M_CSTATE)
    {
        v_stat_0 &= ~M_CSTATE;    /* clear bit. (overwrite) */
    }
    else
    {
        v_stat_0 |= M_CSTATE;    /* set bit. (overwrite) */
    }

    /* fetch x and y coords and complement cursor. */
    //			v9990_SetBackdropColor(i++);
    invert_cell (v_cur_cx, v_cur_cy);
    tBlink = false;
//    }
}

void textbox::doAction (widget::action wAction)
{
    if (wAction == action::Selected)
    {
        if (topScreen->getActiveWidget() != nullptr)
        {
            ((textbox *)topScreen->getActiveWidget())->setActive (false);
        }

        topScreen->setActiveWidget (this);
        ((textbox *)topScreen->getActiveWidget())->setActive (true);
    }
    else
    {
        if (wAction == action::None)
        {
            if (lastAction == action::Selected)
            {
            }
        }
    }

    lastAction = wAction;
}

void textbox::setActive (bool a)
{
    if (a == true)
    {
        cursor_on();
    }
    else
    {
        cursor_off();
    }
}
