/*
 * =====================================================================================
 *
 *       Filename:  textbox.h
 *
 *    Description:  Textbox class
 *
 *        Version:  1.0
 *        Created:  12.05.2016 21:24:21
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Simon Ferber (https://www.ist-schlau.de), simon@ist-schlau.de
 *   Organization:  Kiwi Homebrew Computer
 *
 * =====================================================================================
 */

#ifndef TEXTBOX_H
#define TEXTBOX_H

#include "widget.h"
#include "screen.h"
#include <string>
extern "C" {
#include <v9990.h>
}

class widget;

class textbox : public widget
{
friend class textinput;
public:
    textbox (uint16_t x=0, uint16_t y=0, uint16_t width=0, uint16_t height=0);
    ~textbox();
    void render();
    void edit(uint32_t key);
    void cputc (char c);
    void blink();
    void doAction (widget::action wAction);
    bool tBlink;
private:
    uint16_t tpositionX, tpositionY;
    uint16_t twidth, theight;
    /* VT52 functions & data => new class? */
    uint16_t cursconf (uint16_t, uint16_t);
    void setActive (bool);
    void cell_xfer (uint8_t, uint8_t, uint8_t);
    void neg_cell (uint8_t x, uint8_t y);
    void invert_cell (uint16_t celx, uint16_t cely);
    bool next_cell();
    void move_cursor (int x, int y);
    void cursor_on();
    void cursor_on_cnt();
    void save_cursor_pos();
    void restore_cursor_pos();
    void erase_line();
    void erase_from_bol_impl();
    void erase_from_bol();
    void reverse_video_on();
    void reverse_video_off();
    void set_fg();
    void set_bg();
    void get_fg_col (uint16_t);
    void get_bg_col (uint16_t);
    void erase_from_home();
    void cursor_off();
    void cursor_down_impl();
    void cursor_down();
    void cursor_up();
    void cursor_right();
    void cursor_left_impl();
    void cursor_left();
    void cursor_home();
    void reverse_linefeed();
    void erase_to_eos();
    void erase_to_eol_impl();
    void erase_to_eol();
    void insert_line();
    void delete_line();
    void do_cnt_esce();
    void clear_and_home();
    void ascii_out (int ch);
    void normal_ascii (uint16_t ch);
    void (textbox::*con_state) (uint16_t); /* state of vt52 state machine */
    //typedef void (*con_state_ptr)(uint16_t); /* state of vt52 state machine */
    //con_state_ptr con_state;
    void nop();
    void esc_ch1 (uint16_t);
    void get_row (uint16_t);
    void get_column (uint16_t);
    void blank_out (int topx, int topy, int botx, int boty);
    void scroll_up (int);
    void scroll_down (int);
    void line_wrap_on();
    void line_wrap_off();
    void do_bell();
    void do_backspace();
    void do_tab();
    void ascii_lf();
    void ascii_cr();
    /* jumptable for ASCII control codes */
    void (textbox::*cntl_tab[7]) (void) =
    {
        &textbox::do_bell,            /* 7 = bell */
        &textbox::do_backspace,       /* 8 = backspace */
        &textbox::do_tab,             /* 9 = Horizontal tab */
        &textbox::ascii_lf,           /* 10 = Line feed */
        &textbox::ascii_lf,           /* 11 = Vertical tab (Treated as line feed) */
        &textbox::ascii_lf,           /* 12 = Form Feed (Treated as line feed) */
        &textbox::ascii_cr            /* 13 = Carriage Return */
    };
    /* jumptable for ESC + uppercase character */
    void (textbox::*am_tab[13]) (void) =
    {
        &textbox::cursor_up,          /* Cursor Up */
        &textbox::cursor_down,        /* Cursor Down */
        &textbox::cursor_right,       /* Cursor Right */
        &textbox::cursor_left,        /* Cursor Left */
        &textbox::clear_and_home,     /* Clear and Home */
        &textbox::nop,                /* <ESC> F not supported */
        &textbox::nop,                /* <ESC> G not supported */
        &textbox::cursor_home,        /* Home */
        &textbox::reverse_linefeed,   /* Reverse Line Feed */
        &textbox::erase_to_eos,       /* Erase to End of Screen */
        &textbox::erase_to_eol,       /* Erase to End of Line */
        &textbox::insert_line,        /* Insert Line */
        &textbox::delete_line         /* Delete Line */
    };
    /* jumptable for ESC + lowercase character */
    void (textbox::*bw_tab[22]) (void) =
    {
        &textbox::set_fg,             /* Set foreground color (1 more char) */
        &textbox::set_bg,             /* Set background color (1 more char) */
        &textbox::erase_from_home,    /* Erase from beginning of page */
        &textbox::cursor_on,          /* Cursor On */
        &textbox::cursor_off,         /* Cursor Off */
        &textbox::nop,                /* <ESC> g not supported */
        &textbox::nop,                /* <ESC> h not supported */
        &textbox::nop,                /* <ESC> i not supported */
        &textbox::save_cursor_pos,    /* Save Cursor Position */
        &textbox::restore_cursor_pos, /* Restore Cursor position */
        &textbox::erase_line,         /* Erase line */
        &textbox::nop,                /* <ESC> m not supported */
        &textbox::nop,                /* <ESC> n not supported */
        &textbox::erase_from_bol,     /* Erase from Beginning of Line */
        &textbox::reverse_video_on,   /* Reverse Video On */
        &textbox::reverse_video_off,  /* Reverse Video Off */
        &textbox::nop,                /* <ESC> r not supported */
        &textbox::nop,                /* <ESC> s not supported */
        &textbox::nop,                /* <ESC> t not supported */
        &textbox::nop,                /* <ESC> u not supported */
        &textbox::line_wrap_on,       /* Wrap at End of Line */
        &textbox::line_wrap_off       /* No Wrap at End of Line */
    };
    enum
    {
        M_CFLASH = 0x01, /* cursor flash */
        M_CSTATE = 0x02, /* flash state */
        M_CVIS = 0x04, /* visible */
        M_CEOL = 0x08, /* End of line handling */
        M_REVID = 0x10, /* reverse video */
        M_SVPOS = 0x20, /* position saved */
        M_CRIT = 0x40, /* reverse video */
    };
    uint16_t cols, rows;
    uint16_t save_row;
    uint16_t v_col_bg, v_col_fg;
    uint16_t v_cur_cx, v_cur_cy;
    uint16_t v_cur_of;
    uint8_t v_cur_tim;
    uint8_t v_period;
    uint16_t disab_cnt;
    uint8_t v_stat_0;
    uint16_t sav_cur_x, sav_cur_y;
    uint16_t v_cel_ht, v_cel_mx, v_cel_my, v_cel_wr;
};

#endif

