/*
 * =====================================================================================
 *
 *       Filename:  textinput.cpp
 *
 *    Description:  Textinput class
 *
 *        Version:  1.0
 *        Created:  22.05.2016 15:09:41
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Simon Ferber (https://www.ist-schlau.de), simon@ist-schlau.de
 *   Organization:  Kiwi Homebrew Computer
 *
 * =====================================================================================
 */

#include "textinput.h"

textinput::textinput (uint16_t x, uint16_t y, uint16_t w, uint16_t h) : textbox(x,y,w,h)
{
    buffer = new std::string;
    positionX = x;
    positionY = y;
    width = w;
    height = h;
    widgetType = type::Textinput;
    callback = nullptr;
    tBlink = false;
    /* vt52 init */
    v_cel_ht = 8;
    v_cel_wr = (w>>3);
    v_cel_mx = (w>>3)-1;
    v_cel_my = (h>>3)-1;
    tpositionX = positionX + 2;
    tpositionY = positionY + 2;
    twidth = v_cel_mx << 3; // width - 4;
    theight = v_cel_my << 3; // height - 4;
    /* Initial cursor settings */
    v_cur_cx = 0;                       // cursor to column 0
    v_cur_cy = 0;                       // cursor to line 0
    v_cur_of = 0;                       // line offset is 0
    cursconf (4, 30);                   // .5 second blink rate (@60 Hz vblank).
    cursconf (3, 0);                   // cursor on
    v_cur_tim = v_period;               // load initial value to blink timer
    disab_cnt = 0;                      // cursor disabled 1 level deep.
    v_col_fg = 0x0000;
    v_col_bg = 0x5555;
    con_state = &textinput::normal_ascii;           /* Init conout state machine */
    clear_and_home();
    setActive (false);
    render();
}

textinput::~textinput()
{
}

void textinput::render()
{
    //std::cout << "Ich bin eine Textbox bei x=" << getX() << ", y=" << getY() << std::endl;
    uint16_t color1, color2;
    color1 = 0x6666;
    color2 = 0x6666;
    //color2 = 0x3333;
    static uint16_t color3=0x5555;
    // Draw button background
    VDP_BOX btn_background, caption_box;
    uint16_t btn_bgcolor=color3;//0x2525;
    //color3 += 0x1111;
    btn_background.left = positionX;
    btn_background.top = positionY;
    btn_background.width = width;
    btn_background.height = height;
    vdp_critical |= 0x01;
    v9990_SetCmdWriteMask (0xffff);
    VDPWriteReg (VDP_LOP,VDP_LOP_WCSC);
    v9990_DrawFilledBox (&btn_background, btn_bgcolor++);
    wait_vdp();
    // Draw button border
    v9990_DrawLine (positionX, positionY, width, 0, color1); // top border
    v9990_DrawLine (positionX, positionY, 0, height, color1); // left border
    v9990_DrawLine (positionX, positionY + height, width, 0, color2); // bottom border
    v9990_DrawLine (positionX + width, positionY, 0, height, color2); // right border
    vdp_critical &= ~0x01;
}

std::string textinput::getValue()
{
    return *buffer;
}

void textinput::clearValue()
{
    buffer->clear();
}

void textinput::cputc(char c)
{
    if(c == 0x08) /* ASCII code backspace */
    {
        textbox::cputc(c);
        buffer->erase(v_cur_cx, 1);
        return;

    }
    if(buffer->length() < v_cel_mx) /* Better <= ? */
    {
        *buffer += c;
        textbox::cputc(c);
    }
}
