/*
 * =====================================================================================
 *
 *       Filename:  widget.h
 *
 *    Description:  Widget class
 *
 *        Version:  1.0
 *        Created:  15.05.2016 12:52:25
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Simon Ferber (https://www.ist-schlau.de), simon@ist-schlau.de
 *   Organization:  Kiwi Homebrew Computer
 *
 * =====================================================================================
 */

#ifndef WIDGET_H
#define WIDGET_H

#include <cstdint>
#include <cstdbool>

class screen;

class widget
{
    friend class button;
    friend class textbox;
    friend class textinput;
    friend class label;
    friend class screen;

public:
    enum class action
    {
        None, Show, Hide, Selected
    };

    enum class type
    {
        Button, Textbox, Textinput, Label, Menu
    };
    //widget();
    ~widget();

    void moveXY (uint16_t x, uint16_t y);
    virtual void render (bool pressed=false);
    uint16_t getX();
    uint16_t getY();
    uint16_t getWidth();
    uint16_t getHeight();
    type getType();
    virtual void doAction (action widgetAction);
    virtual action getlastAction();
    virtual void setCallback (void (*cb) (widget *w));
    void setActive (bool);
    bool isActive();
    screen *getTopScreen();
    widget *getActiveWidget();
    uint8_t getWidgetNumber();
    void setWidgetNumber(uint8_t num);
private:
    static uint8_t numWidgets; /* number of widgets */
    uint8_t number; /* this widgets number */
    void (*callback) (widget *w);
    screen *topScreen;
    type widgetType;
    action widgetAction;
    action lastAction;
    uint16_t positionX, positionY;
    uint16_t width, height;
    bool active;
};

#endif
